from .basic_wrappers import clip_reward_v0  # NOQA
from .basic_wrappers import (
    clip_actions_v0,
    color_reduction_v0,
    dtype_v0,
    flatten_v0,
    normalize_obs_v0,
    reshape_v0,
    resize_v1,
    scale_actions_v0,
)
from .delay_observations import delay_observations_v0  # NOQA
from .frame_skip import frame_skip_v0  # NOQA
from .frame_stack import frame_stack_v1, frame_stack_v2  # NOQA
from .max_observation import max_observation_v0  # NOQA
from .nan_wrappers import nan_noop_v0, nan_random_v0, nan_zeros_v0  # NOQA
from .sticky_actions import sticky_actions_v0  # NOQA



import gymnasium
import pytest
from pettingzoo.mpe import simple_spread_v2

from supersuit import concat_vec_envs_v1, pettingzoo_env_to_vec_env_v1
from supersuit.generic_wrappers.frame_skip import frame_skip_gym


def test_env_is_wrapped_true():
    env = gymnasium.make("MountainCarContinuous-v0")
    env = frame_skip_gym(env, 4)
    num_envs = 3
    venv1 = concat_vec_envs_v1(env, num_envs)
    assert venv1.env_is_wrapped(frame_skip_gym) == [True] * 3


def test_env_is_wrapped_false():
    env = gymnasium.make("MountainCarContinuous-v0")
    num_envs = 3
    venv1 = concat_vec_envs_v1(env, num_envs)
    assert venv1.env_is_wrapped(frame_skip_gym) == [False] * 3


@pytest.mark.skip(
    reason="Wrapper depreciated, see https://github.com/Farama-Foundation/SuperSuit/issues/188"
)
def test_env_is_wrapped_cpu():
    env = gymnasium.make("MountainCarContinuous-v0")
    env = frame_skip_gym(env, 4)
    num_envs = 3
    venv1 = concat_vec_envs_v1(env, num_envs, num_cpus=2)
    assert venv1.env_is_wrapped(frame_skip_gym) == [True] * 3


def test_env_is_wrapped_pettingzoo():
    env = simple_spread_v2.parallel_env()
    venv1 = pettingzoo_env_to_vec_env_v1(env)
    num_envs = 3
    venv1 = concat_vec_envs_v1(venv1, num_envs)
    assert venv1.env_is_wrapped(frame_skip_gym) == [False] * 9



import numpy as np
from pettingzoo.butterfly import knights_archers_zombies_v10

from supersuit import vectorize_aec_env_v0


def recursive_equal(info1, info2):
    if info1 == info2:
        return True
    return False


def test_identical():
    def env_fn():
        return knights_archers_zombies_v10.env()  # ,20)

    n_envs = 2
    # single threaded
    env1 = vectorize_aec_env_v0(knights_archers_zombies_v10.env(), n_envs)
    env2 = vectorize_aec_env_v0(knights_archers_zombies_v10.env(), n_envs, num_cpus=1)
    env1.reset(seed=42)
    env2.reset(seed=42)

    def policy(obs, agent):
        return [env1.action_space(agent).sample() for i in range(env1.num_envs)]

    envs_done = 0
    for agent in env1.agent_iter(200000):
        assert env1.agent_selection == env2.agent_selection
        agent = env1.agent_selection
        (
            obs1,
            rew1,
            agent_term1,
            agent_trunc1,
            env_term1,
            env_trunc1,
            agent_passes1,
            infos1,
        ) = env1.last()
        (
            obs2,
            rew2,
            agent_term2,
            agent_trunc2,
            env_term2,
            env_trunc2,
            agent_passes2,
            infos2,
        ) = env2.last()
        assert np.all(np.equal(obs1, obs2))
        assert np.all(np.equal(agent_term1, agent_term2))
        assert np.all(np.equal(agent_trunc1, agent_trunc2))
        assert np.all(np.equal(agent_passes1, agent_passes2))
        assert np.all(np.equal(env_term1, env_term2))
        assert np.all(np.equal(env_trunc1, env_trunc2))
        assert np.all(np.equal(obs1, obs2))
        assert all(
            np.all(np.equal(r1, r2))
            for r1, r2 in zip(env1.rewards.values(), env2.rewards.values())
        )
        assert recursive_equal(infos1, infos2)
        actions = policy(obs1, agent)
        env1.step(actions)
        env2.step(actions)
        # env.envs[0].render()
        for j in range(2):
            # if agent_passes[j]:
            #     print("pass")
            if rew1[j] != 0:
                print(j, agent, rew1, agent_term1[j], agent_trunc1[j])
            if env_term1[j] or env_trunc1[j]:
                print(j, "done")
                envs_done += 1
                if envs_done == n_envs + 1:
                    print("test passed")
                    return



import gymnasium
import numpy as np
from pettingzoo.utils.wrappers import BaseParallelWrapper

from supersuit.utils.wrapper_chooser import WrapperChooser


class black_death_par(BaseParallelWrapper):
    def __init__(self, env):
        super().__init__(env)

    def _check_valid_for_black_death(self):
        for agent in self.agents:
            space = self.observation_space(agent)
            assert isinstance(
                space, gymnasium.spaces.Box
            ), f"observation sapces for black death must be Box spaces, is {space}"

    def reset(self, seed=None, options=None):
        obss = self.env.reset(seed=seed, options=options)

        self.agents = self.env.agents[:]
        self._check_valid_for_black_death()
        black_obs = {
            agent: np.zeros_like(self.observation_space(agent).low)
            for agent in self.agents
            if agent not in obss
        }

        return {**obss, **black_obs}

    def step(self, actions):
        active_actions = {agent: actions[agent] for agent in self.env.agents}
        obss, rews, terms, truncs, infos = self.env.step(active_actions)
        black_obs = {
            agent: np.zeros_like(self.observation_space(agent).low)
            for agent in self.agents
            if agent not in obss
        }
        black_rews = {agent: 0.0 for agent in self.agents if agent not in obss}
        black_infos = {agent: {} for agent in self.agents if agent not in obss}
        terminations = np.fromiter(terms.values(), dtype=bool)
        truncations = np.fromiter(truncs.values(), dtype=bool)
        env_is_done = (terminations & truncations).all()
        total_obs = {**black_obs, **obss}
        total_rews = {**black_rews, **rews}
        total_infos = {**black_infos, **infos}
        total_dones = {agent: env_is_done for agent in self.agents}
        if env_is_done:
            self.agents.clear()
        return total_obs, total_rews, total_dones, total_dones, total_infos


black_death_v3 = WrapperChooser(parallel_wrapper=black_death_par)



from test.dummy_aec_env import DummyEnv

import numpy as np
from gymnasium import spaces

from supersuit import (
    agent_indicator_v0,
    black_death_v3,
    clip_actions_v0,
    clip_reward_v0,
    color_reduction_v0,
    delay_observations_v0,
    dtype_v0,
    flatten_v0,
    frame_skip_v0,
    frame_stack_v1,
    max_observation_v0,
    nan_random_v0,
    nan_zeros_v0,
    normalize_obs_v0,
    pad_action_space_v0,
    pad_observations_v0,
    scale_actions_v0,
    sticky_actions_v0,
)


def observation_homogenizable(env, agents):
    homogenizable = True
    for agent in agents:
        homogenizable = homogenizable and (
            isinstance(env.observation_space(agent), spaces.Box)
            or isinstance(env.observation_space(agent), spaces.Discrete)
        )
    return homogenizable


def action_homogenizable(env, agents):
    homogenizable = True
    for agent in agents:
        homogenizable = homogenizable and (
            isinstance(env.action_space(agent), spaces.Box)
            or isinstance(env.action_space(agent), spaces.Discrete)
        )
    return homogenizable


def image_observation(env, agents):
    imagable = True
    for agent in agents:
        if isinstance(env.observation_space(agent), spaces.Box):
            imagable = imagable and (env.observation_space(agent).low.shape == 3)
            imagable = imagable and (len(env.observation_space(agent).shape[2]) == 3)
            imagable = imagable and (env.observation_space(agent).low == 0).all()
            imagable = imagable and (env.observation_space(agent).high == 255).all()
        else:
            return False
    return imagable


def box_action(env, agents):
    boxable = True
    for agent in agents:
        boxable = boxable and isinstance(env.action_space(agent), spaces.Box)
    return boxable


def not_dict_observation(env, agents):
    is_dict = True
    for agent in agents:
        is_dict = is_dict and (isinstance(env.observation_space(agent), spaces.Dict))
    return not is_dict


def not_discrete_observation(env, agents):
    is_discrete = True
    for agent in agents:
        is_discrete = is_discrete and (
            isinstance(env.observation_space(agent), spaces.Discrete)
        )
    return not is_discrete


def not_multibinary_observation(env, agents):
    is_discrete = True
    for agent in agents:
        is_discrete = is_discrete and (
            isinstance(env.observation_space(agent), spaces.MultiBinary)
        )
    return not is_discrete


def unwrapped_check(env):
    env.reset()
    agents = env.agents

    if image_observation(env, agents):
        env = max_observation_v0(env, 2)
        env = color_reduction_v0(env, mode="full")
        env = normalize_obs_v0(env)

    if box_action(env, agents):
        env = clip_actions_v0(env)
        env = scale_actions_v0(env, 0.5)

    if observation_homogenizable(env, agents):
        env = pad_observations_v0(env)
        env = frame_stack_v1(env, 2)
        env = agent_indicator_v0(env)
        env = black_death_v3(env)

    if (
        not_dict_observation(env, agents)
        and not_discrete_observation(env, agents)
        and not_multibinary_observation(env, agents)
    ):
        env = dtype_v0(env, np.float16)
        env = flatten_v0(env)
        env = frame_skip_v0(env, 2)

    if action_homogenizable(env, agents):
        env = pad_action_space_v0(env)

    env = clip_reward_v0(env, lower_bound=-1, upper_bound=1)
    env = delay_observations_v0(env, 2)
    env = sticky_actions_v0(env, 0.5)
    env = nan_random_v0(env)
    env = nan_zeros_v0(env)

    assert env.unwrapped.__class__ == DummyEnv, f"Failed to unwrap {env}"


def test_unwrapped():
    observation_spaces = []
    base = spaces.Box(low=-1.0, high=1.0, shape=[2], dtype=np.float32)
    observation_spaces.append({f"a{i}": base for i in range(2)})
    base = spaces.Box(low=0, high=255, shape=[64, 64, 3], dtype=np.int8)
    observation_spaces.append({f"a{i}": base for i in range(2)})
    base = spaces.Discrete(5)
    observation_spaces.append({f"a{i}": base for i in range(2)})
    base = spaces.MultiBinary([3, 4])
    observation_spaces.append({f"a{i}": base for i in range(2)})

    action_spaces = []
    base = spaces.Box(-3.0, 3.0, [3], np.float32)
    action_spaces.append({f"a{i}": base for i in range(2)})
    base = spaces.Discrete(5)
    action_spaces.append({f"a{i}": base for i in range(2)})
    base = spaces.MultiDiscrete([4, 5])
    action_spaces.append({f"a{i}": base for i in range(2)})

    for obs_space in observation_spaces:
        for act_space in action_spaces:
            base_obs = {a: obs_space[a].sample() for a in obs_space}
            env = DummyEnv(base_obs, obs_space, act_space)
            unwrapped_check(env)



import random

import numpy as np
from pettingzoo.butterfly import knights_archers_zombies_v10
from pettingzoo.classic import rps_v2
from pettingzoo.mpe import simple_world_comm_v2

from supersuit import vectorize_aec_env_v0


def test_all():
    NUM_ENVS = 5

    def test_vec_env(vec_env):
        vec_env.reset()
        (
            obs,
            rew,
            agent_term,
            agent_trunc,
            env_term,
            env_trunc,
            agent_passes,
            infos,
        ) = vec_env.last()
        print(np.asarray(obs).shape)
        assert len(obs) == NUM_ENVS
        act_space = vec_env.action_space(vec_env.agent_selection)
        assert np.all(np.equal(obs, vec_env.observe(vec_env.agent_selection)))
        assert len(vec_env.observe(vec_env.agent_selection)) == NUM_ENVS
        vec_env.step([act_space.sample() for _ in range(NUM_ENVS)])
        (
            obs,
            rew,
            agent_term,
            agent_trunc,
            env_term,
            env_trunc,
            agent_passes,
            infos,
        ) = vec_env.last(observe=False)
        assert obs is None

    def test_infos(vec_env):
        vec_env.reset()
        infos = vec_env.infos[vec_env.agent_selection]
        assert infos[1]["legal_moves"]

    def test_seed(vec_env):
        vec_env.reset(seed=4)

    def test_some_done(vec_env):
        vec_env.reset()
        act_space = vec_env.action_space(vec_env.agent_selection)
        assert not any(done for dones in vec_env.dones.values() for done in dones)
        vec_env.step([act_space.sample() for _ in range(NUM_ENVS)])
        assert any(rew != 0 for rews in vec_env.rewards.values() for rew in rews)
        any_done_first = any(done for dones in vec_env.dones.values() for done in dones)
        vec_env.step([act_space.sample() for _ in range(NUM_ENVS)])
        any_done_second = any(
            done for dones in vec_env.dones.values() for done in dones
        )
        assert any_done_first and any_done_second

    def select_action(vec_env, passes, i):
        my_info = vec_env.infos[vec_env.agent_selection][i]
        if False and not passes[i] and "legal_moves" in my_info:
            return random.choice(my_info["legal_moves"])
        else:
            act_space = vec_env.action_space(vec_env.agent_selection)
            return act_space.sample()

    for num_cpus in [0, 1]:
        test_vec_env(vectorize_aec_env_v0(rps_v2.env(), NUM_ENVS, num_cpus=num_cpus))
        test_vec_env(
            vectorize_aec_env_v0(
                knights_archers_zombies_v10.env(), NUM_ENVS, num_cpus=num_cpus
            )
        )
        test_vec_env(
            vectorize_aec_env_v0(
                simple_world_comm_v2.env(), NUM_ENVS, num_cpus=num_cpus
            )
        )



from pettingzoo.utils.env import AECEnv, ParallelEnv

from supersuit import action_lambda_v1, observation_lambda_v0
from supersuit.utils.action_transforms import homogenize_ops


def pad_action_space_v0(env):
    assert isinstance(env, AECEnv) or isinstance(
        env, ParallelEnv
    ), "pad_action_space_v0 only accepts an AECEnv or ParallelEnv"
    assert hasattr(
        env, "possible_agents"
    ), "environment passed to pad_observations must have a possible_agents list."
    spaces = [env.action_space(agent) for agent in env.possible_agents]
    homogenize_ops.check_homogenize_spaces(spaces)
    padded_space = homogenize_ops.homogenize_spaces(spaces)
    return action_lambda_v1(
        env,
        lambda action, act_space: homogenize_ops.dehomogenize_actions(
            act_space, action
        ),
        lambda act_space: padded_space,
    )


def pad_observations_v0(env):
    assert isinstance(env, AECEnv) or isinstance(
        env, ParallelEnv
    ), "pad_observations_v0 only accepts an AECEnv or ParallelEnv"
    assert hasattr(
        env, "possible_agents"
    ), "environment passed to pad_observations must have a possible_agents list."
    spaces = [env.observation_space(agent) for agent in env.possible_agents]
    homogenize_ops.check_homogenize_spaces(spaces)
    padded_space = homogenize_ops.homogenize_spaces(spaces)
    return observation_lambda_v0(
        env,
        lambda obs, obs_space: homogenize_ops.homogenize_observations(
            padded_space, obs
        ),
        lambda obs_space: padded_space,
    )



from pettingzoo.butterfly import pistonball_v6

from supersuit import pettingzoo_env_to_vec_env_v1


def make_env():
    env = pistonball_v6.parallel_env()
    env = pettingzoo_env_to_vec_env_v1(env)
    return env


# unfortunately this test does not pass
# def test_vector_render_multiproc():
#     env = make_env()
#     num_envs = 3
#     venv = concat_vec_envs_v1(env, num_envs, num_cpus=num_envs, base_class='stable_baselines3')
#     venv.reset()
#     arr = venv.render(mode="rgb_array")
#     venv.reset()
#     assert len(arr.shape) == 3 and arr.shape[2] == 3
#     venv.reset()
#     try:
#         venv.close()
#     except RuntimeError:
#         pass



from test.dummy_aec_env import DummyEnv

import numpy as np
import pytest
from gymnasium.spaces import Box, Dict, Discrete, Tuple
from gymnasium.vector.utils import concatenate, create_empty_array
from pettingzoo.utils.conversions import aec_to_parallel

from supersuit import concat_vec_envs_v1, pettingzoo_env_to_vec_env_v1

n_agents = 5


def make_env():
    test_env = DummyEnv(
        {
            str(i): {
                "feature": i * np.ones((5,), dtype=np.float32),
                "id": (
                    i * np.ones((7,), dtype=np.float32),
                    i * np.ones((8,), dtype=np.float32),
                ),
            }
            for i in range(n_agents)
        },
        {
            str(i): Dict(
                {
                    "feature": Box(low=0, high=10, shape=(5,)),
                    "id": Tuple(
                        [
                            Box(low=0, high=10, shape=(7,)),
                            Box(low=0, high=10, shape=(8,)),
                        ]
                    ),
                }
            )
            for i in range(n_agents)
        },
        {
            str(i): Dict(
                {
                    "obs": Box(low=0, high=10, shape=(5,)),
                    "mask": Discrete(10),
                }
            )
            for i in range(n_agents)
        },
    )
    test_env.metadata["is_parallelizable"] = True
    return aec_to_parallel(test_env)


def dict_vec_env_test(env):
    # tests that environment really is a vectorized
    # version of the environment returned by make_env

    obss = env.reset()
    for i in range(55):
        actions = [env.action_space.sample() for i in range(env.num_envs)]
        actions = concatenate(
            env.action_space,
            actions,
            create_empty_array(env.action_space, env.num_envs),
        )
        obss, rews, terms, truncs, infos = env.step(actions)
        assert obss["feature"][1][0] == 1
        assert {
            "feature": obss["feature"][1][:],
            "id": [o[1] for o in obss["id"]],
        } in env.observation_space
        # no agent death, only env death
        if any(terms):
            assert all(terms)
        if any(truncs):
            assert all(truncs)


def test_pettingzoo_vec_env():
    env = make_env()
    env = pettingzoo_env_to_vec_env_v1(env)
    dict_vec_env_test(env)


def test_single_threaded_concatenate():
    env = make_env()
    env = pettingzoo_env_to_vec_env_v1(env)
    env = concat_vec_envs_v1(env, 2, num_cpus=1)
    dict_vec_env_test(env)


@pytest.mark.skip(
    reason="Wrapper depreciated, see https://github.com/Farama-Foundation/SuperSuit/issues/188"
)
def test_multi_threaded_concatenate():
    env = make_env()
    env = pettingzoo_env_to_vec_env_v1(env)
    env = concat_vec_envs_v1(env, 2, num_cpus=2)
    dict_vec_env_test(env)


if __name__ == "__main__":
    test_multi_threaded_concatenate()
    exit(0)



from .concat_vec_env import ConcatVecEnv  # NOQA
from .constructors import MakeCPUAsyncConstructor  # NOQA
from .markov_vector_wrapper import MarkovVectorEnv  # NOQA
from .multiproc_vec import ProcConcatVec  # NOQA
from .single_vec_env import SingleVecEnv  # NOQA



import functools

import gymnasium
import numpy as np
from gymnasium.spaces import Box, Space

from supersuit.utils.base_aec_wrapper import BaseWrapper
from supersuit.utils.wrapper_chooser import WrapperChooser


class aec_observation_lambda(BaseWrapper):
    def __init__(self, env, change_observation_fn, change_obs_space_fn=None):
        assert callable(
            change_observation_fn
        ), "change_observation_fn needs to be a function. It is {}".format(
            change_observation_fn
        )
        assert change_obs_space_fn is None or callable(
            change_obs_space_fn
        ), "change_obs_space_fn needs to be a function. It is {}".format(
            change_obs_space_fn
        )

        self.change_observation_fn = change_observation_fn
        self.change_obs_space_fn = change_obs_space_fn

        super().__init__(env)

        if hasattr(self, "possible_agents"):
            for agent in self.possible_agents:
                # call any validation logic in this function
                self.observation_space(agent)

    def _modify_action(self, agent, action):
        return action

    def _check_wrapper_params(self):
        if self.change_obs_space_fn is None and hasattr(self, "possible_agents"):
            for agent in self.possible_agents:
                assert isinstance(
                    self.observation_space(agent), Box
                ), "the observation_lambda_wrapper only allows the change_obs_space_fn argument to be optional for Box observation spaces"

    @functools.lru_cache(maxsize=None)
    def observation_space(self, agent):
        if self.change_obs_space_fn is None:
            space = self.env.observation_space(agent)
            try:
                trans_low = self.change_observation_fn(space.low, space, agent)
                trans_high = self.change_observation_fn(space.high, space, agent)
            except TypeError:
                trans_low = self.change_observation_fn(space.low, space)
                trans_high = self.change_observation_fn(space.high, space)
            new_low = np.minimum(trans_low, trans_high)
            new_high = np.maximum(trans_low, trans_high)

            return Box(low=new_low, high=new_high, dtype=new_low.dtype)
        else:
            old_obs_space = self.env.observation_space(agent)
            try:
                return self.change_obs_space_fn(old_obs_space, agent)
            except TypeError:
                return self.change_obs_space_fn(old_obs_space)

    def _modify_observation(self, agent, observation):
        old_obs_space = self.env.observation_space(agent)
        try:
            return self.change_observation_fn(observation, old_obs_space, agent)
        except TypeError:
            return self.change_observation_fn(observation, old_obs_space)


class gym_observation_lambda(gymnasium.Wrapper):
    def __init__(self, env, change_observation_fn, change_obs_space_fn=None):
        assert callable(
            change_observation_fn
        ), "change_observation_fn needs to be a function. It is {}".format(
            change_observation_fn
        )
        assert change_obs_space_fn is None or callable(
            change_obs_space_fn
        ), "change_obs_space_fn needs to be a function. It is {}".format(
            change_obs_space_fn
        )
        self.change_observation_fn = change_observation_fn
        self.change_obs_space_fn = change_obs_space_fn

        super().__init__(env)
        self._check_wrapper_params()
        self._modify_spaces()

    def _check_wrapper_params(self):
        if self.change_obs_space_fn is None:
            space = self.observation_space
            assert isinstance(
                space, Box
            ), "the observation_lambda_wrapper only allows the change_obs_space_fn argument to be optional for Box observation spaces"

    def _modify_spaces(self):
        space = self.observation_space

        if self.change_obs_space_fn is None:
            new_low = self.change_observation_fn(space.low, space)
            new_high = self.change_observation_fn(space.high, space)
            new_space = Box(low=new_low, high=new_high, dtype=new_low.dtype)
        else:
            new_space = self.change_obs_space_fn(space)
            assert isinstance(
                new_space, Space
            ), "output of change_obs_space_fn to observation_lambda_wrapper must be a gymnasium space"
        self.observation_space = new_space

    def _modify_observation(self, observation):
        return self.change_observation_fn(observation, self.env.observation_space)

    def step(self, action):
        observation, rew, termination, truncation, info = self.env.step(action)
        observation = self._modify_observation(observation)
        return observation, rew, termination, truncation, info

    def reset(self, seed=None, options=None):
        observation = self.env.reset(seed=seed, options=options)
        observation = self._modify_observation(observation)
        return observation


observation_lambda_v0 = WrapperChooser(
    aec_wrapper=aec_observation_lambda, gym_wrapper=gym_observation_lambda
)



from supersuit.generic_wrappers import clip_actions_v0  # NOQA
from supersuit.generic_wrappers import (
    clip_reward_v0,
    color_reduction_v0,
    delay_observations_v0,
    dtype_v0,
    flatten_v0,
    frame_skip_v0,
    frame_stack_v1,
    max_observation_v0,
    normalize_obs_v0,
    reshape_v0,
    resize_v1,
    sticky_actions_v0,
)

from .aec_vector import vectorize_aec_env_v0
from .generic_wrappers import *  # NOQA
from .lambda_wrappers import observation_lambda_v0  # NOQA
from .lambda_wrappers import action_lambda_v1, reward_lambda_v0
from .multiagent_wrappers import black_death_v3  # NOQA
from .multiagent_wrappers import (
    agent_indicator_v0,
    pad_action_space_v0,
    pad_observations_v0,
)
from .vector.vector_constructors import concat_vec_envs_v1  # NOQA
from .vector.vector_constructors import (
    gym_vec_env_v0,
    pettingzoo_env_to_vec_env_v1,
    stable_baselines3_vec_env_v0,
    stable_baselines_vec_env_v0,
)


class DeprecatedWrapper(ImportError):
    pass


def __getattr__(wrapper_name):
    """
    Gives error that looks like this when trying to import old version of wrapper:
    File "./supersuit/__init__.py", line 38, in __getattr__
    raise DeprecatedWrapper(f"{base}{version_num} is now deprecated, use {base}{act_version_num} instead")
    supersuit.DeprecatedWrapper: concat_vec_envs_v0 is now deprecated, use concat_vec_envs_v1 instead
    """
    start_v = wrapper_name.rfind("_v") + 2
    version = wrapper_name[start_v:]
    base = wrapper_name[:start_v]
    try:
        version_num = int(version)
        is_valid_version = True
    except ValueError:
        is_valid_version = False

    globs = globals()
    if is_valid_version:
        for act_version_num in range(1000):
            if f"{base}{act_version_num}" in globs:
                if version_num < act_version_num:
                    raise DeprecatedWrapper(
                        f"{base}{version_num} is now deprecated, use {base}{act_version_num} instead"
                    )

    raise ImportError(f"cannot import name '{wrapper_name}' from 'supersuit'")


__version__ = "3.7.2"



import numpy as np

from . import convert_box

COLOR_RED_LIST = ["full", "R", "G", "B"]
GRAYSCALE_WEIGHTS = np.array([0.299, 0.587, 0.114], dtype=np.float32)


def check_param(space, color_reduction):
    assert isinstance(
        color_reduction, str
    ), f"color_reduction must be str. It is {color_reduction}"
    assert color_reduction in COLOR_RED_LIST, "color_reduction must be in {}".format(
        COLOR_RED_LIST
    )
    assert (
        len(space.low.shape) == 3 and space.low.shape[2] == 3
    ), "To apply color_reduction, shape must be a 3d image with last dimension of size 3. Shape is {}".format(
        space.low.shape
    )


def change_obs_space(obs_space, param):
    return convert_box(lambda obs: change_observation(obs, obs_space, param), obs_space)


def change_observation(obs, obs_space, color_reduction):
    if color_reduction == "R":
        obs = obs[:, :, 0]
    if color_reduction == "G":
        obs = obs[:, :, 1]
    if color_reduction == "B":
        obs = obs[:, :, 2]
    if color_reduction == "full":
        obs = (obs.astype(np.float32) @ GRAYSCALE_WEIGHTS).astype(np.uint8)
    return obs



import copy

import gymnasium
import numpy as np
import pytest
from pettingzoo.mpe import simple_spread_v2

from supersuit import concat_vec_envs_v1, gym_vec_env_v0, pettingzoo_env_to_vec_env_v1


def recursive_equal(info1, info2):
    try:
        if info1 == info2:
            return True
    except ValueError:
        if isinstance(info1, np.ndarray) and isinstance(info2, np.ndarray):
            return np.all(np.equal(info1, info2))
        elif isinstance(info1, dict) and isinstance(info2, dict):
            return all(
                (
                    set(info1.keys()) == set(info2.keys())
                    and recursive_equal(info1[i], info2[i])
                )
                for i in info1.keys()
            )
        elif isinstance(info1, list) and isinstance(info2, list):
            return all(recursive_equal(i1, i2) for i1, i2 in zip(info1, info2))
    return False


def check_vec_env_equivalency(venv1, venv2, check_info=True):
    # assert venv1.observation_space == venv2.observation_space
    # assert venv1.action_space == venv2.action_space

    obs1 = venv1.reset(seed=51)
    obs2 = venv2.reset(seed=51)

    for i in range(400):
        action = [venv1.action_space.sample() for env in range(venv1.num_envs)]
        assert np.all(np.equal(obs1, obs2))

        obs1, rew1, term1, trunc1, info1 = venv1.step(action)
        obs2, rew2, term2, trunc2, info2 = venv2.step(action)

        # uses close rather than equal due to inconsistency in reporting rewards as float32 or float64
        assert np.allclose(rew1, rew2)
        assert np.all(np.equal(term1, term2))
        assert np.all(np.equal(trunc1, trunc2))
        assert recursive_equal(info1, info2) or not check_info


@pytest.mark.skip(
    reason="Wrapper depreciated, see https://github.com/Farama-Foundation/SuperSuit/issues/188"
)
def test_gym_supersuit_equivalency():
    env = gymnasium.make("MountainCarContinuous-v0")
    num_envs = 3
    venv1 = concat_vec_envs_v1(env, num_envs)
    venv2 = gym_vec_env_v0(env, num_envs)
    check_vec_env_equivalency(venv1, venv2)


@pytest.mark.skip(
    reason="Wrapper depreciated, see https://github.com/Farama-Foundation/SuperSuit/issues/188"
)
def test_inital_state_dissimilarity():
    env = gymnasium.make("CartPole-v1")
    venv = concat_vec_envs_v1(env, 2)
    observations = venv.reset()
    assert not np.equal(observations[0], observations[1]).all()


# we really don't want to have a stable baselines dependency even in tests
# def test_stable_baselines_supersuit_equivalency():
#     env = gymnasium.make("MountainCarContinuous-v0")
#     num_envs = 3
#     venv1 = supersuit_vec_env(env, num_envs, base_class='stable_baselines3')
#     venv2 = stable_baselines3_vec_env(env, num_envs)
#     check_vec_env_equivalency(venv1, venv2, check_info=False)  # stable baselines does not implement info correctly


@pytest.mark.skip(
    reason="Wrapper depreciated, see https://github.com/Farama-Foundation/SuperSuit/issues/188"
)
def test_mutliproc_single_proc_equivalency():
    env = gymnasium.make("CartPole-v1")
    num_envs = 3
    # uses single threaded vector environment
    venv1 = concat_vec_envs_v1(env, num_envs, num_cpus=0)
    # uses multiprocessing vector environment
    venv2 = concat_vec_envs_v1(env, num_envs, num_cpus=4)
    check_vec_env_equivalency(venv1, venv2)


@pytest.mark.skip(
    reason="Wrapper depreciated, see https://github.com/Farama-Foundation/SuperSuit/issues/188"
)
def test_multiagent_mutliproc_single_proc_equivalency():
    env = simple_spread_v2.parallel_env(max_cycles=10)
    env = pettingzoo_env_to_vec_env_v1(env)
    num_envs = 3
    # uses single threaded vector environment
    venv1 = concat_vec_envs_v1(env, num_envs, num_cpus=0)
    # uses multiprocessing vector environment
    venv2 = concat_vec_envs_v1(env, num_envs, num_cpus=4)
    check_vec_env_equivalency(venv1, venv2)


@pytest.mark.skip(
    reason="Wrapper depreciated, see https://github.com/Farama-Foundation/SuperSuit/issues/188"
)
def test_multiproc_buffer():
    num_envs = 2
    env = gymnasium.make("CartPole-v1")
    env = concat_vec_envs_v1(env, num_envs, num_cpus=2)

    obss = env.reset()
    for i in range(55):
        actions = [env.action_space.sample() for i in range(env.num_envs)]

        # Check we're not passing a thing that gets mutated
        keep_obs = copy.deepcopy(obss)
        new_obss, rews, terms, truncs, infos = env.step(actions)

        assert hash(str(keep_obs)) == hash(str(obss))

        obss = new_obss



from test.dummy_gym_env import DummyEnv

import numpy as np
from gymnasium import spaces

from supersuit import (
    clip_actions_v0,
    clip_reward_v0,
    color_reduction_v0,
    delay_observations_v0,
    dtype_v0,
    flatten_v0,
    frame_skip_v0,
    frame_stack_v1,
    max_observation_v0,
    nan_random_v0,
    nan_zeros_v0,
    normalize_obs_v0,
    scale_actions_v0,
    sticky_actions_v0,
)


def unwrapped_check(env):
    # image observations
    if isinstance(env.observation_space, spaces.Box):
        if (
            (env.observation_space.low.shape == 3)
            and (env.observation_space.low == 0).all()
            and (len(env.observation_space.shape[2]) == 3)
            and (env.observation_space.high == 255).all()
        ):
            env = max_observation_v0(env, 2)
            env = color_reduction_v0(env, mode="full")
            env = normalize_obs_v0(env)

    # box action spaces
    if isinstance(env.action_space, spaces.Box):
        env = clip_actions_v0(env)
        env = scale_actions_v0(env, 0.5)

    # stackable observations
    if isinstance(env.observation_space, spaces.Box) or isinstance(
        env.observation_space, spaces.Discrete
    ):
        env = frame_stack_v1(env, 2)

    # not discrete and not multibinary observations
    if not isinstance(env.observation_space, spaces.Discrete) and not isinstance(
        env.observation_space, spaces.MultiBinary
    ):
        env = dtype_v0(env, np.float16)
        env = flatten_v0(env)
        env = frame_skip_v0(env, 2)

    # everything else
    env = clip_reward_v0(env, lower_bound=-1, upper_bound=1)
    env = delay_observations_v0(env, 2)
    env = sticky_actions_v0(env, 0.5)
    env = nan_random_v0(env)
    env = nan_zeros_v0(env)

    assert env.unwrapped.__class__ == DummyEnv, f"Failed to unwrap {env}"


def test_unwrapped():
    observation_spaces = []
    observation_spaces.append(
        spaces.Box(low=-1.0, high=1.0, shape=[2], dtype=np.float32)
    )
    observation_spaces.append(
        spaces.Box(low=0, high=255, shape=[64, 64, 3], dtype=np.int8)
    )
    observation_spaces.append(spaces.Discrete(5))
    observation_spaces.append(spaces.MultiBinary([3, 4]))

    action_spaces = []
    action_spaces.append(spaces.Box(-3.0, 3.0, [3], np.float32))
    action_spaces.append(spaces.Discrete(5))
    action_spaces.append(spaces.MultiDiscrete([3, 5]))

    for obs_space in observation_spaces:
        for act_space in action_spaces:
            env = DummyEnv(obs_space.sample(), obs_space, act_space)
            unwrapped_check(env)



import numpy as np
from pettingzoo.utils.agent_selector import agent_selector

from .base_aec_vec_env import VectorAECEnv


class SyncAECVectorEnv(VectorAECEnv):
    def __init__(self, env_constructors):
        assert len(env_constructors) >= 1
        assert callable(
            env_constructors[0]
        ), "env_constructor must be a callable object (i.e function) that create an environment"

        self.envs = [env_constructor() for env_constructor in env_constructors]
        self.num_envs = len(env_constructors)
        self.env = self.envs[0]
        self.max_num_agents = self.env.max_num_agents
        self.possible_agents = self.env.possible_agents
        self._agent_selector = agent_selector(self.possible_agents)

    def action_space(self, agent):
        return self.env.action_space(agent)

    def observation_space(self, agent):
        return self.env.observation_space(agent)

    def _find_active_agent(self):
        cur_selection = self.agent_selection
        while not any(cur_selection == env.agent_selection for env in self.envs):
            cur_selection = self._agent_selector.next()
        return cur_selection

    def _collect_dicts(self):
        self.rewards = {
            agent: np.array(
                [
                    env.rewards[agent] if agent in env.rewards else 0
                    for env in self.envs
                ],
                dtype=np.float32,
            )
            for agent in self.possible_agents
        }
        self._cumulative_rewards = {
            agent: np.array(
                [
                    env._cumulative_rewards[agent]
                    if agent in env._cumulative_rewards
                    else 0
                    for env in self.envs
                ],
                dtype=np.float32,
            )
            for agent in self.possible_agents
        }
        self.terminations = {
            agent: np.array(
                [
                    env.terminations[agent] if agent in env.terminations else True
                    for env in self.envs
                ],
                dtype=np.uint8,
            )
            for agent in self.possible_agents
        }
        self.truncations = {
            agent: np.array(
                [
                    env.truncations[agent] if agent in env.truncations else True
                    for env in self.envs
                ],
                dtype=np.uint8,
            )
            for agent in self.possible_agents
        }
        self.infos = {
            agent: [env.infos[agent] if agent in env.infos else {} for env in self.envs]
            for agent in self.possible_agents
        }

    def reset(self, seed=None, options=None):
        """
        returns: list of observations
        """
        if seed is not None:
            for i, env in enumerate(self.envs):
                env.reset(seed=seed + i, options=options)
        else:
            for i, env in enumerate(self.envs):
                env.reset(seed=None, options=options)

        self.agent_selection = self._agent_selector.reset()
        self.agent_selection = self._find_active_agent()

        self._collect_dicts()
        self.envs_terminations = np.zeros(self.num_envs)
        self.envs_truncations = np.zeros(self.num_envs)

    def observe(self, agent):
        observations = []
        for env in self.envs:
            obs = (
                env.observe(agent)
                if (agent in env.terminations) or (agent in env.truncations)
                else np.zeros_like(env.observation_space(agent).low)
            )
            observations.append(obs)
        return np.stack(observations)

    def last(self, observe=True):
        passes = np.array(
            [env.agent_selection != self.agent_selection for env in self.envs],
            dtype=np.uint8,
        )
        last_agent = self.agent_selection
        obs = self.observe(last_agent) if observe else None
        return (
            obs,
            self._cumulative_rewards[last_agent],
            self.terminations[last_agent],
            self.truncations[last_agent],
            self.envs_terminations,
            self.envs_truncations,
            passes,
            self.infos[last_agent],
        )

    def step(self, actions, observe=True):
        assert len(actions) == len(
            self.envs
        ), f"{len(actions)} actions given, but there are {len(self.envs)} environments!"
        old_agent = self.agent_selection

        envs_dones = []
        for i, (act, env) in enumerate(zip(actions, self.envs)):
            # Prior to the truncation API update, the env was reset if env.agents was an empty list
            # After the truncation API update, the env needs to be reset if every agent is terminated OR truncated
            terminations = np.fromiter(env.terminations.values(), dtype=bool)
            truncations = np.fromiter(env.truncations.values(), dtype=bool)
            env_done = (terminations | truncations).all()
            envs_dones.append(env_done)

            if env_done:
                env.reset()
            elif env.agent_selection == old_agent:
                if isinstance(type(act), np.ndarray):
                    act = np.array(act)
                act = (
                    act
                    if not (
                        self.terminations[old_agent][i]
                        or self.truncations[old_agent][i]
                    )
                    else None
                )  # if the agent is dead, set action to None
                env.step(act)

        self.agent_selection = self._agent_selector.next()
        self.agent_selection = self._find_active_agent()

        self.envs_dones = np.array(envs_dones)
        self._collect_dicts()



import numpy as np
from gymnasium import spaces
from gymnasium.spaces import Box, Discrete


def check_homogenize_spaces(all_spaces):
    assert len(all_spaces) > 0
    space1 = all_spaces[0]
    assert all(
        isinstance(space, space1.__class__) for space in all_spaces
    ), "all spaces to homogenize must be of same general shape"

    if isinstance(space1, spaces.Box):
        for space in all_spaces:
            assert isinstance(
                space, spaces.Box
            ), "all spaces for homogenize must be either Box or Discrete, not a mix"
            assert len(space1.shape) == len(
                space.shape
            ), "all spaces to homogenize must be of same shape"
            assert (
                space1.dtype == space.dtype
            ), "all spaces to homogenize must be of same dtype"
    elif isinstance(space1, spaces.Discrete):
        for space in all_spaces:
            assert isinstance(
                space, spaces.Discrete
            ), "all spaces for homogenize must be either Box or Discrete, not a mix"
    else:
        assert False, "homogenization only supports Discrete and Box spaces"


def pad_to(arr, new_shape, pad_value):
    old_shape = arr.shape
    if old_shape == new_shape:
        return arr
    pad_size = [ns - os for ns, os in zip(new_shape, old_shape)]
    pad_tuples = [(0, ps) for ps in pad_size]
    return np.pad(arr, pad_tuples, constant_values=pad_value)


def homogenize_spaces(all_spaces):
    space1 = all_spaces[0]
    if isinstance(space1, spaces.Box):
        all_dims = np.array([space.shape for space in all_spaces], dtype=np.int32)
        max_dims = np.max(all_dims, axis=0)
        new_shape = tuple(max_dims)
        all_lows = np.stack(
            [
                pad_to(space.low, new_shape, np.minimum(0, np.min(space.low)))
                for space in all_spaces
            ]
        )
        all_highs = np.stack(
            [
                pad_to(space.high, new_shape, np.maximum(1e-5, np.max(space.high)))
                for space in all_spaces
            ]
        )
        new_low = np.min(all_lows, axis=0)
        new_high = np.max(all_highs, axis=0)
        assert new_shape == new_low.shape
        return Box(low=new_low, high=new_high, dtype=space1.dtype)
    elif isinstance(space1, spaces.Discrete):
        max_n = max([space.n for space in all_spaces])
        return Discrete(max_n)
    else:
        assert False


def dehomogenize_actions(orig_action_space, action):
    if isinstance(orig_action_space, spaces.Box):
        # choose only the relevant action values
        cur_shape = action.shape
        new_shape = orig_action_space.shape
        if cur_shape == new_shape:
            return action
        else:
            assert len(cur_shape) == len(new_shape)
            slices = [slice(0, i) for i in new_shape]
            new_action = action[tuple(slices)]

            return new_action

    elif isinstance(orig_action_space, spaces.Discrete):
        # extra action values refer to action value 0
        n = orig_action_space.n
        if action is None:
            return None
        if action > n - 1:
            action = 0
        return action
    else:
        assert False


def homogenize_observations(obs_space, obs):
    if isinstance(obs_space, spaces.Box):
        return pad_to(obs, obs_space.shape, 0)
    elif isinstance(obs_space, spaces.Discrete):
        return obs_space
    else:
        assert False



from pettingzoo import AECEnv
from pettingzoo.utils.agent_selector import agent_selector


class DummyEnv(AECEnv):
    metadata = {"render_modes": ["human"], "is_parallelizable": True}

    def __init__(self, observations, observation_spaces, action_spaces):
        super().__init__()
        self._observations = observations
        self._observation_spaces = observation_spaces

        self.agents = sorted([x for x in observation_spaces.keys()])
        self.possible_agents = self.agents[:]
        self._agent_selector = agent_selector(self.agents)
        self.agent_selection = self._agent_selector.reset()
        self._action_spaces = action_spaces

        self.steps = 0

    def observation_space(self, agent):
        return self._observation_spaces[agent]

    def action_space(self, agent):
        return self._action_spaces[agent]

    def observe(self, agent):
        return self._observations[agent]

    def step(self, action, observe=True):
        if (
            self.terminations[self.agent_selection]
            or self.truncations[self.agent_selection]
        ):
            return self._was_dead_step(action)
        self._cumulative_rewards[self.agent_selection] = 0
        self.agent_selection = self._agent_selector.next()
        self.steps += 1
        if self.steps >= 5 * len(self.agents):
            self.truncations = {a: True for a in self.agents}

        self._accumulate_rewards()
        self._deads_step_first()

    def reset(self, seed=None, options=None):
        self.agents = self.possible_agents[:]
        self._agent_selector = agent_selector(self.agents)
        self.agent_selection = self._agent_selector.reset()
        self.rewards = {a: 1 for a in self.agents}
        self._cumulative_rewards = {a: 0 for a in self.agents}
        self.terminations = {a: False for a in self.agents}
        self.truncations = {a: False for a in self.agents}
        self.infos = {a: {} for a in self.agents}
        self.steps = 0

    def close(self):
        pass



"""Setups up the SuperSuit module."""

from setuptools import find_packages, setup


def get_description():
    """Gets the description from the readme."""
    with open("README.md") as fh:
        long_description = ""
        header_count = 0
        for line in fh:
            if line.startswith("##"):
                header_count += 1
            if header_count < 2:
                long_description += line
            else:
                break
    return header_count, long_description


def get_version():
    """Gets the pettingzoo version."""
    path = "supersuit/__init__.py"
    with open(path) as file:
        lines = file.readlines()

    for line in lines:
        if line.startswith("__version__"):
            return line.strip().split()[-1].strip().strip('"')
    raise RuntimeError("bad version data in __init__.py")


version = get_version()
header_count, long_description = get_description()

setup(
    name="SuperSuit",
    version=version,
    author="Farama Foundation",
    author_email="contact@farama.org",
    description="Wrappers for Gymnasium and PettingZoo",
    url="https://github.com/Farama-Foundation/SuperSuit",
    license_files=("LICENSE.txt",),
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=["Reinforcement Learning", "game", "RL", "AI", "gymnasium"],
    python_requires=">=3.7, <3.12",
    packages=find_packages(),
    install_requires=["numpy>=1.19.0", "gymnasium>=0.26.0"],
    extras={"dev": ["pettingzoo[butterfly]"]},
    classifiers=[
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
)



import copy

import pytest
from pettingzoo.butterfly import knights_archers_zombies_v10
from pettingzoo.mpe import simple_spread_v2, simple_world_comm_v2

from supersuit import black_death_v3, concat_vec_envs_v1, pettingzoo_env_to_vec_env_v1


def test_good_env():
    env = simple_spread_v2.parallel_env()
    max_num_agents = len(env.possible_agents)
    env = pettingzoo_env_to_vec_env_v1(env)
    assert env.num_envs == max_num_agents

    obss = env.reset()
    for i in range(55):
        actions = [env.action_space.sample() for i in range(env.num_envs)]

        # Check we're not passing a thing that gets mutated
        keep_obs = copy.deepcopy(obss)
        new_obss, rews, terms, truncs, infos = env.step(actions)

        assert hash(str(keep_obs)) == hash(str(obss))
        assert len(new_obss) == max_num_agents
        assert len(rews) == max_num_agents
        assert len(terms) == max_num_agents
        assert len(truncs) == max_num_agents
        assert len(infos) == max_num_agents
        # no agent death, only env death
        if any(terms):
            assert all(terms)
        if any(truncs):
            assert all(truncs)
        obss = new_obss


def test_good_vecenv():
    num_envs = 2
    env = simple_spread_v2.parallel_env()
    max_num_agents = len(env.possible_agents) * num_envs
    env = pettingzoo_env_to_vec_env_v1(env)
    env = concat_vec_envs_v1(env, num_envs)

    obss = env.reset()
    for i in range(55):
        actions = [env.action_space.sample() for i in range(env.num_envs)]

        # Check we're not passing a thing that gets mutated
        keep_obs = copy.deepcopy(obss)
        new_obss, rews, terms, truncs, infos = env.step(actions)

        assert hash(str(keep_obs)) == hash(str(obss))
        assert len(new_obss) == max_num_agents
        assert len(rews) == max_num_agents
        assert len(terms) == max_num_agents
        assert len(truncs) == max_num_agents
        assert len(infos) == max_num_agents
        # no agent death, only env death
        if any(terms):
            assert all(terms)
        if any(truncs):
            assert all(truncs)
        obss = new_obss


def test_bad_action_spaces_env():
    env = simple_world_comm_v2.parallel_env()
    with pytest.raises(AssertionError):
        env = pettingzoo_env_to_vec_env_v1(env)


def test_env_black_death_assertion():
    env = knights_archers_zombies_v10.parallel_env(spawn_rate=50, max_cycles=2000)
    env = pettingzoo_env_to_vec_env_v1(env)
    with pytest.raises(AssertionError):
        for i in range(100):
            env.reset()
            for i in range(2000):
                actions = [env.action_space.sample() for i in range(env.num_envs)]
                obss, rews, terms, truncs, infos = env.step(actions)


def test_env_black_death_wrapper():
    env = knights_archers_zombies_v10.parallel_env(spawn_rate=50, max_cycles=300)
    env = black_death_v3(env)
    env = pettingzoo_env_to_vec_env_v1(env)
    env.reset()
    for i in range(300):
        actions = [env.action_space.sample() for i in range(env.num_envs)]
        obss, rews, terms, truncs, infos = env.step(actions)



from . import convert_box


def check_param(obs_space, should_flatten):
    assert isinstance(
        should_flatten, bool
    ), f"should_flatten must be bool. It is {should_flatten}"


def change_obs_space(obs_space, param):
    return convert_box(lambda obs: change_observation(obs, obs_space, param), obs_space)


def change_observation(obs, obs_space, should_flatten):
    if should_flatten:
        obs = obs.flatten()
    return obs



import numpy as np
import pytest
from pettingzoo.test import api_test, parallel_test
from pettingzoo.test.example_envs import (
    generated_agents_env_v0,
    generated_agents_parallel_v0,
)

import supersuit
from supersuit import dtype_v0

wrappers = [
    supersuit.dtype_v0(generated_agents_parallel_v0.env(), np.int32),
    supersuit.flatten_v0(generated_agents_parallel_v0.env()),
    supersuit.normalize_obs_v0(
        dtype_v0(generated_agents_parallel_v0.env(), np.float32),
        env_min=-1,
        env_max=5.0,
    ),
    supersuit.frame_stack_v1(generated_agents_parallel_v0.env(), 8),
    supersuit.reward_lambda_v0(generated_agents_parallel_v0.env(), lambda x: x / 10),
    supersuit.clip_reward_v0(generated_agents_parallel_v0.env()),
    supersuit.nan_noop_v0(generated_agents_parallel_v0.env(), 0),
    supersuit.nan_zeros_v0(generated_agents_parallel_v0.env()),
    supersuit.nan_random_v0(generated_agents_parallel_v0.env()),
    supersuit.frame_skip_v0(generated_agents_parallel_v0.env(), 4),
    supersuit.sticky_actions_v0(generated_agents_parallel_v0.env(), 0.75),
    supersuit.delay_observations_v0(generated_agents_parallel_v0.env(), 3),
    supersuit.max_observation_v0(generated_agents_parallel_v0.env(), 3),
]


@pytest.mark.parametrize("env", wrappers)
def test_pettingzoo_aec_api_par_gen(env):
    api_test(env, num_cycles=50)


wrappers = [
    supersuit.dtype_v0(generated_agents_env_v0.env(), np.int32),
    supersuit.flatten_v0(generated_agents_env_v0.env()),
    supersuit.normalize_obs_v0(
        dtype_v0(generated_agents_env_v0.env(), np.float32), env_min=-1, env_max=5.0
    ),
    supersuit.frame_stack_v1(generated_agents_env_v0.env(), 8),
    supersuit.reward_lambda_v0(generated_agents_env_v0.env(), lambda x: x / 10),
    supersuit.clip_reward_v0(generated_agents_env_v0.env()),
    supersuit.nan_noop_v0(generated_agents_env_v0.env(), 0),
    supersuit.nan_zeros_v0(generated_agents_env_v0.env()),
    supersuit.nan_random_v0(generated_agents_env_v0.env()),
    supersuit.frame_skip_v0(generated_agents_env_v0.env(), 4),
    supersuit.sticky_actions_v0(generated_agents_env_v0.env(), 0.75),
    supersuit.delay_observations_v0(generated_agents_env_v0.env(), 3),
    supersuit.max_observation_v0(generated_agents_env_v0.env(), 3),
]


@pytest.mark.parametrize("env", wrappers)
def test_pettingzoo_aec_api_aec_gen(env):
    api_test(env, num_cycles=50)


parallel_wrappers = wrappers = [
    supersuit.dtype_v0(generated_agents_parallel_v0.parallel_env(), np.int32),
    supersuit.flatten_v0(generated_agents_parallel_v0.parallel_env()),
    supersuit.normalize_obs_v0(
        dtype_v0(generated_agents_parallel_v0.parallel_env(), np.float32),
        env_min=-1,
        env_max=5.0,
    ),
    supersuit.frame_stack_v1(generated_agents_parallel_v0.parallel_env(), 8),
    supersuit.reward_lambda_v0(
        generated_agents_parallel_v0.parallel_env(), lambda x: x / 10
    ),
    supersuit.clip_reward_v0(generated_agents_parallel_v0.parallel_env()),
    supersuit.nan_noop_v0(generated_agents_parallel_v0.parallel_env(), 0),
    supersuit.nan_zeros_v0(generated_agents_parallel_v0.parallel_env()),
    supersuit.nan_random_v0(generated_agents_parallel_v0.parallel_env()),
    supersuit.frame_skip_v0(generated_agents_parallel_v0.parallel_env(), 4, 0),
    supersuit.sticky_actions_v0(generated_agents_parallel_v0.parallel_env(), 0.75),
    supersuit.delay_observations_v0(generated_agents_parallel_v0.parallel_env(), 3),
    supersuit.max_observation_v0(generated_agents_parallel_v0.parallel_env(), 3),
]


@pytest.mark.parametrize("env", parallel_wrappers)
def test_pettingzoo_parallel_api_gen(env):
    parallel_test.parallel_api_test(env, num_cycles=50)


wrapper_fns = [
    lambda: supersuit.pad_action_space_v0(generated_agents_parallel_v0.env()),
    lambda: supersuit.pad_observations_v0(generated_agents_parallel_v0.env()),
    lambda: supersuit.agent_indicator_v0(generated_agents_parallel_v0.env()),
    lambda: supersuit.vectorize_aec_env_v0(generated_agents_parallel_v0.env(), 2),
    lambda: supersuit.pad_action_space_v0(generated_agents_parallel_v0.parallel_env()),
    lambda: supersuit.pad_observations_v0(generated_agents_parallel_v0.parallel_env()),
    lambda: supersuit.agent_indicator_v0(generated_agents_parallel_v0.parallel_env()),
    lambda: supersuit.pettingzoo_env_to_vec_env_v1(
        generated_agents_parallel_v0.parallel_env()
    ),
]


@pytest.mark.parametrize("wrapper_fn", wrapper_fns)
def test_pettingzoo_missing_optional_error_message(wrapper_fn):
    with pytest.raises(AssertionError, match=" must have "):
        wrapper_fn()



import gymnasium
from pettingzoo.utils.conversions import aec_to_parallel, parallel_to_aec
from pettingzoo.utils.env import AECEnv, ParallelEnv


class WrapperChooser:
    def __init__(self, aec_wrapper=None, gym_wrapper=None, parallel_wrapper=None):
        assert (
            aec_wrapper is not None or parallel_wrapper is not None
        ), "either the aec wrapper or the parallel wrapper must be defined for all supersuit environments"
        self.aec_wrapper = aec_wrapper
        self.gym_wrapper = gym_wrapper
        self.parallel_wrapper = parallel_wrapper

    def __call__(self, env, *args, **kwargs):
        if isinstance(env, gymnasium.Env):
            if self.gym_wrapper is None:
                raise ValueError(
                    f"{self.wrapper_name} does not apply to gymnasium environments, pettingzoo environments only"
                )
            return self.gym_wrapper(env, *args, **kwargs)
        elif isinstance(env, AECEnv):
            if self.aec_wrapper is not None:
                return self.aec_wrapper(env, *args, **kwargs)
            else:
                return parallel_to_aec(
                    self.parallel_wrapper(aec_to_parallel(env), *args, **kwargs)
                )
        elif isinstance(env, ParallelEnv):
            if self.parallel_wrapper is not None:
                return self.parallel_wrapper(env, *args, **kwargs)
            else:
                return aec_to_parallel(
                    self.aec_wrapper(parallel_to_aec(env), *args, **kwargs)
                )
        else:
            raise ValueError(
                "environment passed to supersuit wrapper must either be a gymnasium environment or a pettingzoo environment"
            )
